# This file is part of ubuntu-bug-triage. See LICENSE file for license info.
# pylint: disable=redefined-outer-name, no-member
"""Test view module."""
from mock import Mock
import pytest

from .bug import Bug
from .test_bug import MockBug1234567
from .view import TerminalView, CSVView, JSONView, BrowserView


@pytest.fixture(scope="module")
def bugs():
    """Create array of bugs for testing."""
    return [Bug(MockBug1234567()), Bug(MockBug1234567())]


def test_browser_view(bugs):
    """Test out browser is launched."""
    # mock out the sleeps to prevent slowing down unit tests
    import time

    time.sleep = Mock()
    time.sleep.return_value = None

    # mock out webbrowser open and open tab functions
    import webbrowser

    webbrowser.open = Mock()
    webbrowser.open.return_value = True
    webbrowser.open_new_tab = Mock()
    webbrowser.open_new_tab.return_value = True

    assert BrowserView(bugs)
    assert webbrowser.open.called
    assert webbrowser.open_new_tab.called


def test_csv_view(bugs):
    """Test CSV view."""
    assert CSVView(bugs)


def test_json_view(bugs):
    """Test JSON view."""
    assert JSONView(bugs)


def test_terminal_view(bugs):
    """Test terminal view."""
    # Mock out the terminal width call
    import os

    os.popen = Mock()
    os.popen.return_value.read.return_value.split.return_value = "80"

    assert TerminalView(bugs)



# This file is part of ubuntu-bug-triage. See LICENSE file for license info.
"""Bug module."""


class BugTask:
    """Bug Task class."""

    def __init__(self, lp_link):
        """Initialize Bug Task."""
        self._lp = lp_link

        self.src_pkg = self.target.split(" ")[0]

    def __repr__(self):
        """Representation of specific task."""
        return "- %-20s [%s]" % (self.target, self.status)

    @property
    def importance(self):
        """Return importance of bug task."""
        return self._lp.importance

    @property
    def status(self):
        """Return status of bug task."""
        return self._lp.status

    @property
    def target(self):
        """Return target of bug task."""
        return self._lp.bug_target_name


class Bug:
    """Bug class."""

    def __init__(self, lp_link, ignore_user=[]):
        """Initialize bug class."""
        self._lp = lp_link

        self.id = self._lp.id
        self.title = self._lp.title

        self.tasks = []

        for task in self._lp.bug_tasks:
            self.tasks.append(BugTask(task))

        if not ignore_user:
            self.last_active_user = ""
        else:
            for comment in self._lp.messages:
                datecreated = comment.date_created
                commenter = comment.owner_link.split("~")[1]

            for active in self._lp.activity:
                datechanged = active.datechanged
                person = active.person_link.split("~")[1]

            if datecreated > datechanged:
                self.last_active_user = commenter
            else:
                self.last_active_user = person

    def __eq__(self, other):
        """Two bugs are identical if their ids are equal."""
        return self.id == other.id

    @property
    def affects(self):
        """Create string of all packages affected."""
        return [task.target for task in self.tasks]

    @property
    def url(self):
        """Return URL of bug."""
        return self._lp.web_link

    def to_json(self):
        """Return JSON representation of object."""
        return {"id": self.id, "title": self.title, "affects": self.affects}



# This file is part of ubuntu-bug-triage. See LICENSE file for license info.
"""Triage module."""

import itertools
import logging
import os

from launchpadlib.launchpad import Launchpad
from launchpadlib.credentials import UnencryptedFileCredentialStore

from . import BLACKLIST
from .bug import Bug


class Triage:
    """Base triage class."""

    def __init__(self, date, anon):
        """Initialize triage class."""
        self._log = logging.getLogger(__name__)

        self.launchpad = self._launchpad_connect(anon)
        self.date = date

    def current_backlog_count(self):
        """Return the total current backlog count."""
        raise NotImplementedError

    def updated_bugs(self):
        """Return updated bugs."""
        raise NotImplementedError

    def _launchpad_connect(self, anon=False):
        """Use the launchpad module connect to launchpad.

        Will connect you to the Launchpad website the first time you
        run this to authorize your system to connect unless anonymous
        login is specified.
        """
        if anon:
            self._log.debug("logging into Launchpad anonymously")
            return Launchpad.login_anonymously(
                "ubuntu-bug-triage", "production", version="devel"
            )

        self._log.debug("logging into Launchpad")
        credential_store = UnencryptedFileCredentialStore(
            os.path.expanduser("~/.lp_creds")
        )
        return Launchpad.login_with(
            "ubuntu-bug-triage",
            "production",
            version="devel",
            credential_store=credential_store,
        )

    @staticmethod
    def _tasks_to_bug_ids(tasks):
        """Take list of tasks and return unique set of bug ids."""
        bugs = []
        for task in tasks:
            bug_id = task.bug_link.split("/")[-1]
            if bug_id not in bugs:
                bugs.append(bug_id)

        return sorted(bugs)


class TeamTriage(Triage):
    """Triage Launchpad bugs for a particular Ubuntu team."""

    def __init__(self, team, date, anon, status, ignore_user):
        """Initialize Team Triage."""
        super().__init__(date, anon)

        self._log.debug("finding bugs for team: %s", team)
        self.team = self.launchpad.people[team]
        self.status = status
        self.ignore_user = ignore_user

    def current_backlog_count(self):
        """Get team's current backlog count."""
        return len(
            self.launchpad.distributions["Ubuntu"].searchTasks(
                bug_subscriber=self.team, status=self.status
            )
        )

    def updated_bugs(self):
        """Print update bugs for a specific date or date range."""
        updated_tasks = self.launchpad.distributions["Ubuntu"].searchTasks(
            modified_since=self.date,
            structural_subscriber=self.team,
            status=self.status,
        )

        bugs = []
        for bug_id in sorted(self._tasks_to_bug_ids(updated_tasks)):
            bug = Bug(self.launchpad.bugs[bug_id], self.ignore_user)

            if self.team.name in BLACKLIST:
                if self._all_src_on_blacklist(bug.tasks, self.team.name):
                    self._log.debug("skipping bug: %s", bug_id)
                    continue

            if bug.last_active_user not in self.ignore_user:
                bugs.append(bug)

        return bugs

    @staticmethod
    def _all_src_on_blacklist(tasks, team):
        """Test if bug tasks source packages are all on blacklist."""
        for task in tasks:
            if task.src_pkg not in BLACKLIST[team]:
                return False

        return True


class PackageTriage(Triage):
    """Triage Launchpad bugs for a particular package."""

    def __init__(self, package, date, anon, include_project, status, ignore_user):
        """Initialize package triage."""
        super().__init__(date, anon)

        self._log.debug("finding bugs for package: %s", package)
        self.package = self.launchpad.distributions["Ubuntu"].getSourcePackage(
            name=package
        )
        self.ignore_user = ignore_user

        if self.package is None and not include_project:
            self._log.warning("warn: no Ubuntu package with that name exists")

        self.project = None
        if include_project:
            try:
                self.project = self.launchpad.projects[package]
            except KeyError:
                self._log.warning("warn: no Launchpad project with that name exists")

        self.status = status

    def current_backlog_count(self):
        """Get packages's current backlog count."""
        count = 0
        if self.package:
            count += len(self.package.searchTasks(status=self.status))
        if self.project:
            count += len(self.project.searchTasks(status=self.status))
        return count

    def updated_bugs(self):
        """Print update bugs for a specific date or date range."""
        package_tasks = []
        if self.package is not None:
            package_tasks = self.package.searchTasks(
                modified_since=self.date, status=self.status
            )
        project_tasks = []
        if self.project is not None:
            project_tasks = self.project.searchTasks(
                modified_since=self.date, status=self.status
            )

        # launchpadlib Collections don't support appending one another, so
        # synthesise an iterable containing both the Collections we care about
        updated_tasks = itertools.chain(package_tasks, project_tasks)

        bugs = []
        for bug_id in sorted(self._tasks_to_bug_ids(updated_tasks)):
            bug = Bug(self.launchpad.bugs[bug_id], self.ignore_user)
            if bug.last_active_user not in self.ignore_user:
                bugs.append(bug)

        return bugs



# This file is part of ubuntu-bug-triage. See LICENSE file for license info.
"""Ubuntu Bug Triage module."""

import argparse
from datetime import datetime, timedelta
import logging
import sys

from . import UBUNTU_PACKAGE_TEAMS
from . import ACTIONABLE_BUG_STATUSES
from .triage import PackageTriage, TeamTriage
from .view import BrowserView, CSVView, JSONView, TerminalView


def parse_args():
    """Set up command-line arguments."""
    parser = argparse.ArgumentParser("ubuntu-bug-triage")
    parser.add_argument(
        "package_or_team",
        nargs="?",
        default="ubuntu-server",
        help="""source package name (e.g. cloud-init, lxd) or package team name
        (e.g. ubuntu-openstack, foundations-bugs) to use for search (default:
        ubuntu-server)""",
    )
    parser.add_argument(
        "days",
        nargs="?",
        type=int,
        default=1,
        help="""number of days (e.g 1, 10) of bugs to triage, where one day is
        the default. Use --start-time to use a specific date.""",
    )
    parser.add_argument(
        "--anon", action="store_true", help="Anonymous login to Launchpad"
    )
    parser.add_argument("--csv", action="store_true", help="output as CSV")
    parser.add_argument(
        "--debug", action="store_true", help="additional logging output"
    )
    parser.add_argument(
        "--ignore-user",
        default=[],
        nargs="*",
        help="""ignore bugs edited last by the listed person""",
    )
    parser.add_argument("--json", action="store_true", help="output as JSON")
    parser.add_argument(
        "--open", action="store_true", help="open resulting bugs in web browser"
    )
    parser.add_argument(
        "--include-project",
        "-p",
        action="store_true",
        help="include project bugs in output",
    )
    parser.add_argument(
        "--start-time",
        help="""a date and time (e.g. '2020-10-01 14:15') to start triage
        from. This will override the value set by days.""",
    )
    parser.add_argument(
        "--status",
        "-s",
        action="append",
        default=[],
        metavar="STATUS",
        choices=[
            "any",
            "New",
            "Incomplete",
            "Opinion",
            "Invalid",
            "Won't Fix",
            "Expired",
            "Confirmed",
            "Triaged",
            "In Progress",
            "Fix Committed",
            "Fix Released",
            "Incomplete (with response)",
            "Incomplete (without response)",
        ],
        help="Restrict the search to bugs with the given status."
        " Can be specified multiple times. Defaults: "
        + ", ".join(ACTIONABLE_BUG_STATUSES)
        + ".",
    )
    parser.add_argument(
        "--urls", action="store_true", help="print only the urls of bugs to triage"
    )

    return parser.parse_args()


def parse_date(args):
    """Parse either the number of days for triage or specific date time."""
    if args.start_time:
        try:
            date = datetime.strptime(args.start_time, "%Y-%m-%d %H:%M")
        except ValueError:
            print("Oops: invalid format or date for start time. Use '%Y-%m-%d %H:%M'")
            sys.exit(1)
    else:
        date = datetime.now() - timedelta(days=args.days)

    return date.strftime("%Y-%m-%dT%H:%M")


def setup_logging(debug):
    """Set up logging."""
    logging.basicConfig(
        stream=sys.stdout,
        format="%(message)s",
        level=logging.DEBUG if debug else logging.INFO,
    )


def launch():
    """Launch ubuntu-bug-triage."""
    args = parse_args()
    if not args.status:
        args.status = ACTIONABLE_BUG_STATUSES
    elif "any" in args.status:
        args.status = []
    args.status = list(set(args.status))
    setup_logging(args.debug)
    date = parse_date(args)

    if args.package_or_team in UBUNTU_PACKAGE_TEAMS:
        if args.include_project:
            logging.getLogger(__name__).warning(
                "N.B. --include-project has no effect when running against a"
                " package team"
            )
        triage = TeamTriage(
            args.package_or_team, date, args.anon, args.status, args.ignore_user
        )
    else:
        triage = PackageTriage(
            args.package_or_team,
            date,
            args.anon,
            args.include_project,
            args.status,
            args.ignore_user,
        )

    bugs = triage.updated_bugs()
    if args.urls:
        for bug in bugs:
            print(bug.url)
    elif args.csv:
        CSVView(bugs)
    elif args.json:
        JSONView(bugs)
    else:
        TerminalView(bugs)

    if args.open:
        BrowserView(bugs)


if __name__ == "__main__":
    sys.exit(launch())



# This file is part of ubuntu-bug-triage. See LICENSE file for license info.
"""Test bug module."""
import random

from .bug import Bug, BugTask


class MockBugTaskRandom:
    """Create a task with random valid options."""

    def __init__(self):
        """Initialize class."""
        self.importance = random.choice(
            ["Wishlist", "Low", "Medium", "High", "Critical", "Undecided"]
        )
        self.status = random.choice(
            [
                "New",
                "Incomplete",
                "Opinion",
                "Invalid",
                "Won't Fix",
                "Confirmed",
                "Triaged",
                "In Progress",
                "Fix Committed",
                "Fix Released",
            ]
        )
        self.bug_target_name = random.choice(["Trusty", "Xenial", "Bionic", "Disco"])


class MockBugTask:
    """Create mock bug task with specific items."""

    def __init__(self):
        """Initialize class."""
        self.importance = "Wishlist"
        self.status = "Triaged"
        self.bug_target_name = "Bionic"


class MockBug:
    """Create fake bug with random tasks."""

    def __init__(self):
        """Initialize class."""
        self.id = "1761240"
        self.title = "SRU pollinate 4.33"
        self.status = "Fix Released"
        self.importance = "Low"
        self.bug_target_name = "pollinate (Ubuntu)"
        self.web_link = (
            "https://bugs.launchpad.net/ubuntu/+source/pollinate/+bug/1761240"
        )
        num_tasks = range(random.randint(1, 8))
        self.bug_tasks = [MockBugTaskRandom() for _ in num_tasks]


class MockBugNoTasks:
    """Create bug with no tasks."""

    def __init__(self):
        """Initialize class."""
        self.id = "1761240"
        self.title = "SRU pollinate 4.33"
        self.status = "Fix Released"
        self.importance = "Low"
        self.bug_target_name = "pollinate (Ubuntu)"
        self.web_link = (
            "https://bugs.launchpad.net/ubuntu/+source/pollinate/+bug/1761240"
        )
        self.bug_tasks = []


class MockBug1234567:
    """Fake bug with specific data."""

    def __init__(self):
        """Initialize class."""
        self.id = "1234567"
        self.title = "Test Bug"
        self.status = "In Progress"
        self.importance = "Critical"
        self.bug_target_name = "cloud-init (Ubuntu)"
        self.web_link = (
            "https://bugs.launchpad.net/ubuntu/+source/cloud-init/+bug/1234567"
        )
        self.bug_tasks = []


class MockBug1111111:
    """Another fake bug with specific data."""

    def __init__(self):
        """Initialize class."""
        self.id = "1111111"
        self.title = "Test Bug 2"
        self.status = "Triaged"
        self.importance = "Wishlist"
        self.bug_target_name = "curtin (Ubuntu)"
        self.web_link = (
            "https://bugs.launchpad.net/ubuntu/+source/cloud-init/+bug/1111111"
        )
        self.bug_tasks = []


def test_bug_class():
    """Test Bug Class."""
    assert Bug(MockBug())


def test_bug_class_no_tasks():
    """Test Bug Class with no tasks."""
    assert Bug(MockBugNoTasks())


def test_bug_same():
    """Verify bug compare works."""
    assert Bug(MockBug1234567()) == Bug(MockBug1234567())


def test_bug_not_same():
    """Verify bug compare works negative case."""
    assert Bug(MockBug1234567()) != Bug(MockBug1111111())


def test_affects():
    """Verify affects is an array."""
    bug = Bug(MockBug())
    assert isinstance(bug.affects, type([]))


def test_no_affects():
    """Verify no affects."""
    bug = Bug(MockBugNoTasks())
    assert isinstance(bug.affects, type([]))
    assert bug.affects == []


def test_url():
    """Verify url is a string."""
    bug = Bug(MockBug())
    assert isinstance(bug.url, str)


def test_json():
    """Verify JSON output."""
    bug = Bug(MockBug1234567())
    json_output = {"id": bug.id, "title": bug.title, "affects": bug.affects}
    assert bug.to_json() == json_output


def test_bug_task():
    """Verify bug task attributes."""
    task = BugTask(MockBugTask())
    assert task.importance == "Wishlist"
    assert task.status == "Triaged"
    assert repr(task) == "- %-20s [%s]" % (task.target, task.status)



# This file is part of ubuntu-bug-triage. See LICENSE file for license info.
"""Ubuntu Bug Triage module."""

# This is a hard-coded list of teams that produce a lot of bugs and may
# not want to be included in the full team triage.
#
BLACKLIST = {
    "ubuntu-server": [
        "cloud-init",
        "curtin",
        "juju",
        "juju-core",
        "lxc",
        "lxd",
        "maas",
    ]
}

# This is hard-coded as it took way too long to download the json file
# from the following URL:
#
# https://people.canonical.com/~ubuntu-archive/package-team-mapping.json
#
UBUNTU_PACKAGE_TEAMS = [
    "pkg-ime",
    "ubuntu-openstack",
    "documentation-packages",
    "kubuntu-bugs",
    "checkbox-bugs",
    "foundations-bugs",
    "maas-maintainers",
    "ubuntu-server",
    "ubuntu-printing",
    "ubuntu-security",
    "mir-team",
    "kernel-packages",
    "unsubscribed",
    "desktop-packages",
    "translators-packages",
    "snappy-dev",
]

# By default show only bugs with any of the following statuses.
ACTIONABLE_BUG_STATUSES = [
    "New",
    "Incomplete",
    "Confirmed",
    "Triaged",
    "In Progress",
]



"""Setup.py file."""

import os
from setuptools import find_packages, setup

PWD = os.path.abspath(os.path.dirname(__name__))
REQUIREMENTS_FILE = os.path.join(PWD, "requirements.txt")
REQUIREMENTS = []
with open(REQUIREMENTS_FILE, "r") as req_file:
    REQUIREMENTS = req_file.read().splitlines()

README_FILE = os.path.join(PWD, "README.md")
with open(README_FILE, "r") as readme:
    README_TEXT = readme.read()

setup(
    name="ubuntu-bug-triage",
    version="21.1",
    description="Recently updated Launchpad bugs for triage",
    long_description=README_TEXT,
    long_description_content_type="text/markdown",
    author="Joshua Powers",
    author_email="josh.powers@canonical.com",
    url="https://github.com/powersj/ubuntu-bug-triage",
    download_url=("https://github.com/powersj/ubuntu-bug-triage/tarball/master"),
    install_requires=REQUIREMENTS,
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Quality Assurance",
        "Topic :: Software Development :: Testing",
    ],
    keywords=["ubuntu", "launchpad", "triage", "bugs"],
    packages=find_packages(),
    entry_points={
        "console_scripts": ["ubuntu-bug-triage=ubuntu_bug_triage.__main__:launch"]
    },
    zip_safe=True,
)



# This file is part of ubuntu-bug-triage. See LICENSE file for license info.
# flake8: ignore=F811
"""Test triage module."""
import pytest
from mock import Mock

from .triage import Triage


def test_launchpad_login():
    """Test launchpad login."""
    from launchpadlib.credentials import UnencryptedFileCredentialStore

    UnencryptedFileCredentialStore = Mock()  # noqa: F811
    UnencryptedFileCredentialStore.return_value = True

    from launchpadlib.launchpad import Launchpad

    Launchpad.login_with = Mock()
    Launchpad.login_with.return_value = True

    test_triage = Triage(1, False)
    assert test_triage

    with pytest.raises(NotImplementedError):
        test_triage.current_backlog_count()

    with pytest.raises(NotImplementedError):
        test_triage.updated_bugs()


def test_launchpad_login_anon():
    """Test anonymous login."""
    from launchpadlib.credentials import UnencryptedFileCredentialStore

    UnencryptedFileCredentialStore = Mock()  # noqa: F811
    UnencryptedFileCredentialStore.return_value = True

    from launchpadlib.launchpad import Launchpad

    Launchpad.login_anonymously = Mock()
    Launchpad.login_anonymously.return_value = True

    assert Triage(1, True)



# This file is part of ubuntu-bug-triage. See LICENSE file for license info.
"""Triage module."""

import json
import logging
import shutil
import textwrap
import time
import webbrowser

from tabulate import tabulate


class BaseView:
    """Base view class."""

    def __init__(self):
        """Initialize base view."""
        self._log = logging.getLogger(__name__)


class BrowserView(BaseView):
    """Browser view class.

    This class uses the webbrowser from the Python standard library.
    There is no way to know if a page has fully loaded or know when
    a browser is even open. As such this is fire and forget behavior
    and it leads to errors on opening pages.

    I looked into selenium, but found mixed results and had to figure
    out what browser to use in the first place and went with the crude
    sleep statements below.
    """

    def __init__(self, bugs):
        """Initialize browser view."""
        super().__init__()

        self._log.debug("opening bugs in browser")
        self.opened = False
        for bug in bugs:
            self._open_url(bug.url)

    def _open_url(self, url):
        """Open a given url."""
        if self.opened:
            webbrowser.open_new_tab(url)
            time.sleep(1.2)
        else:
            webbrowser.open(url)
            self.opened = True
            time.sleep(5)


class CSVView(BaseView):
    """CSV view class."""

    def __init__(self, bugs):
        """Initialize CSV view."""
        super().__init__()

        print("id,affects,title")
        for bug in bugs:
            print('LP: #%s,"%s",%s' % (bug.id, ",".join(bug.affects), bug.title))


class JSONView(BaseView):
    """JSON view class."""

    def __init__(self, bugs):
        """Initialize JSON view."""
        super().__init__()

        print(json.dumps([bug.to_json() for bug in bugs], indent=4, ensure_ascii=False))


class TerminalView(BaseView):
    """Terminal view class."""

    def __init__(self, bugs):
        """Initialize terminal view."""
        super().__init__()

        # shutil.get_terminal_size() fallback default is 80x24.
        self.term_columns = shutil.get_terminal_size().columns

        # The table has a set number characters that always exist:
        #     4 for '|' boarders
        #     6 for spaces around content
        #    12 for 'LP: #1234567'
        width_table = 22

        # This seemed liked the best middle of the road value
        width_affects = 20

        # The title width will expand as the terminal expands,
        # but will expect a width no less than 80
        width_title = max(80, int(self.term_columns)) - width_affects - width_table

        table = []
        for bug in bugs:
            affects = [
                textwrap.fill(
                    pkg.replace("(Ubuntu)", "(U)").replace("(Debian)", "(D)"),
                    width=width_affects,
                    subsequent_indent="    ",
                )
                for pkg in bug.affects
            ]

            table.append(
                [
                    "LP: #%s" % bug.id,
                    "\n".join(affects),
                    textwrap.fill(bug.title, width=width_title, max_lines=4),
                ]
            )

        print(tabulate(table, ["id", "affects", "title"], "grid"))
